/* 
 * Graph Bisect Library
 * Implementation header
 *
 * Grid Challenge 2006 Final
 * Team Leo - Nan DUN
 * Department of Computer Science, the University of Tokyo
 */

#ifndef GB_IMPL_H
#define	GB_IMPL_H

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <limits.h>
#include <sys/stat.h>
#include "gb.h"

#define GB_ERR_MEM	1	/* Memory error */
#define	GB_ERR_OVER	2	/* Overflow */
#define	GB_ERR_NULL	3	/* Null pointer */
#define	GB_ERR_FILE	4	/* File error */
#define GB_ERR_ARG	5	/* Wrong argument */

#define THRESHOLD	0

#define IDVL_RATIO 10
#define	TABU_NUM 25
#define VOTE_RATIO 1000.0
#define SCORING_TRIALS	3

#define LOCAL_SEARCH
#define LOCAL_SEARCH_NUM 50

typedef struct item {
	integer index;
	struct item *prev, *next;
} item_t;

typedef struct bucket {
	struct item *head;
	struct item *tail;
} bucket_t;

typedef struct itemp {
	integer value;	
	struct item *ptr;
} itemp_t;

typedef struct elite {
	integer edgecut;
	integer *assign;
} elite_t;

typedef struct pair {
	integer a;
	integer b;
} pair_t;
/********************************************
 * Global variables
 *******************************************/
clock_t g_io_start, g_io_end, g_bi_start, g_bi_end, g_timeout, g_timelimit;
integer *g_nv, *g_ne, ***g_adjlist, *g_vparity, *g_maxvdeg, *g_minvdeg, **g_belonglist, *g_edgecut;
integer g_seedL, g_seedU, g_iter;
GB_ALGO g_algo;
integer g_bestedgecut, *g_bestbelonglist, *g_randomvlist;
integer g_threshold;
int g_exitflag;



void GB_Error(int errno, const char *FUNCNAME);
integer GB_Random(integer start, integer step);
void GB_RandomPermute(integer *list, integer n);
void GB_RandomPermuteExclude(integer *list, integer n, integer ex1, integer ex2);
void GB_BucketInitial(bucket_t **buckets, integer nb, itemp_t **items, integer ni, integer value);
void GB_BucketFinalize(bucket_t **buckets, itemp_t **items, integer ni);
void GB_BucketReset(bucket_t *buckets, integer nb, itemp_t *items, integer ni, integer value);
void GB_BucketInsert(integer value, integer index, bucket_t *buckets, integer *min, integer *max, itemp_t *items);
void GB_BucketDelete(integer value, integer index, bucket_t *buckets, integer *min, integer *max, itemp_t *items);
integer GB_BucketGetFirstMin(bucket_t *buckets, integer min);
integer GB_BucketGetFirstMax(bucket_t *buckets, integer max);

void GB_Metis2Johnson(const char *filefrom, const char *fileto);
void GB_DimacsB2Johnson(const char *filefrom, const char *fileto);
void GB_ReadGraph(const char *filename, GB_GRAPH *graph);
void GB_FreeGraph(GB_GRAPH *graph);
void GB_PrintGraph(GB_GRAPH *graph);

void GB_GlobalInitial(GB_GRAPH *graph, GB_ALGO options, integer *seed, int iter, double timeout, integer threshold);
void GB_GlobalFinalized();

void GB_KernighanLinBisect();
void GB_KernighanLinBisectSub();

void GB_LinearKernighanLinBisect();
void GB_LinearKernighanLinBisectSub();

void GB_GreedyGrowingRegionBisect();
void GB_GreedyGrowingRegionBisectSub();

void GB_MinMaxGreedyBisect();
void GB_MinMaxGreedyBisectSub();

void GB_RandomizedTabuSearchBisect();
void GB_ReactiveRandomizedTSBisect();

#endif /* GB_IMPL_H */

