/* 
 * Graph Bisect Library
 * Global Routines
 *
 * Grid Challenge 2006 Final
 * Team Leo - Nan DUN
 * Department of Computer Science, the University of Tokyo
 */

#include "gbimpl.h"

extern integer *g_dvaluelist, *g_edgecutlist;
extern int *g_vlocklist;
extern pair_t *g_vpairlist;
extern bucket_t *g_connBuckets[3];
extern integer g_connMin[3], g_connMax[3];
extern itemp_t *g_vconnlist[3];
extern bucket_t *g_gainBuckets;
extern integer g_gainMin[3], g_gainMax[3], g_gainBktsOffset;
extern itemp_t *g_vgainlist;
extern integer *g_vcutlist;
extern integer g_idvl, g_time, g_time_best, *g_lastusedlist;
extern double g_vote[TABU_NUM + 1];
extern elite_t g_elite[TABU_NUM + 1];
extern integer *g_vinedgecnt, *g_voutedgecnt, *g_lr, *g_rl, *g_ls;
extern bucket_t *g_gainBkt[3];
extern integer g_gainBktMin[3], g_gainBktMax[3], g_gainBktOffset;
extern itemp_t *g_gainlst[3];

extern void ConnBucketInitial(bucket_t **buckets, integer nb, itemp_t **vlist, integer ni);
extern void ConnBucketFinalize(bucket_t **buckets, itemp_t **vlist, integer ni);
extern void GainBucketInitial(bucket_t **buckets, integer ng, itemp_t **vlist, integer ni);
extern void GainBucketFinalize(bucket_t **buckets, itemp_t **vlist, integer ni);

void GB_GlobalInitial(GB_GRAPH *graph, GB_ALGO options, integer *seed, int iter, double timeout, integer threshold) {
	const char FUNCNAME[] = "GB_GlobalInitial(4)";
	
	integer i;

	if (graph == NULL) {
		GB_Error(GB_ERR_NULL, FUNCNAME);
	}
	
	g_nv = &(graph->nv);
    g_ne = &(graph->ne);
    g_adjlist = &(graph->adjlist);
	g_vparity = &(graph->vparity);
    g_maxvdeg = &(graph->maxvdeg);
	g_minvdeg = &(graph->minvdeg);
    g_belonglist = &(graph->belonglist);
    g_edgecut = &(graph->edgecut);
	
	g_algo = options;

	if (seed == NULL) {
		g_seedL = 1;
		g_seedU = *g_nv;
	} else {
		g_seedL = seed[0];
		g_seedU = seed[1];
	
	}

	if (iter == -1) {
		g_iter = INT_MAX;
	} else {
		g_iter = iter;
	}

	g_timeout = (long)(timeout * CLOCKS_PER_SEC);
	g_threshold = threshold;
	g_exitflag = GB_FALSE;
	
    g_randomvlist = (integer *) malloc((*g_nv + 1) * sizeof(integer));
	if (g_randomvlist == NULL) {
		GB_Error(GB_ERR_MEM, FUNCNAME);
	}

	g_bestedgecut = *g_edgecut;
	g_bestbelonglist = (integer *) malloc((*g_nv + 1) * sizeof(integer));
	if (g_bestbelonglist == NULL) {
		GB_Error(GB_ERR_MEM, FUNCNAME);
	}
	for (i = 1; i <= *g_nv; i++) {
		g_bestbelonglist[i] = (*g_belonglist)[i];
	}

	if (g_algo == GB_KL) {
		g_dvaluelist = (integer *) malloc((*g_nv + 1) * sizeof(integer));
		g_vlocklist = (int *) malloc((*g_nv + 1) * sizeof(int));
		g_edgecutlist = (integer *) malloc((*g_nv / 2 + 1) * sizeof(integer));
		g_vpairlist = (pair_t *) malloc((*g_nv / 2 + 1) * sizeof(pair_t));
		if (g_dvaluelist == NULL || g_vlocklist == NULL ||
			g_edgecutlist == NULL || g_vpairlist == NULL) {
			GB_Error(GB_ERR_MEM, FUNCNAME);
		}
	}
	
	if (g_algo == GB_LKL || g_algo == GB_RRTS) {
		g_vlocklist = (int *) malloc((*g_nv + 1) * sizeof(int));
		g_edgecutlist = (integer *) malloc((*g_nv / 2 + 1) * sizeof(integer));
		g_vpairlist = (pair_t *) malloc((*g_nv / 2 + 1) * sizeof(pair_t));
		if (g_vlocklist == NULL || g_edgecutlist == NULL || g_vpairlist == NULL) {
			GB_Error(GB_ERR_MEM, FUNCNAME);
		}
		g_gainBkt[0] = NULL;
		GainBucketInitial(&(g_gainBkt[GB_PAR_L]), 2 * (*g_maxvdeg) + 1, &(g_gainlst[GB_PAR_L]), *g_nv + 1);
		GainBucketInitial(&(g_gainBkt[GB_PAR_R]), 2 * (*g_maxvdeg) + 1, &(g_gainlst[GB_PAR_R]), *g_nv + 1);
		g_gainBktMin[GB_PAR_L] = g_gainBktMin[GB_PAR_R] = *g_maxvdeg;
		g_gainBktMax[GB_PAR_L] = g_gainBktMax[GB_PAR_R] = 0 - *g_maxvdeg;
		g_gainBktOffset = *g_maxvdeg;
	}
	if (g_algo == GB_GG) {
		GainBucketInitial(&g_gainBuckets, 2 * (*g_maxvdeg) + 1, &g_vgainlist, *g_nv + 1);
		g_gainMin[0] = *g_maxvdeg;
		g_gainMax[0] = 0 - *g_maxvdeg;
		g_gainBktsOffset = *g_maxvdeg;
		g_vcutlist = (integer *) malloc((*g_nv + 1) * sizeof(integer));
		if (g_vcutlist == NULL) {
			GB_Error(GB_ERR_MEM, FUNCNAME);
		}
	}

	if (g_algo == GB_MMG || g_algo == GB_RTS || g_algo == GB_RRTS) {
		g_connBuckets[0] = g_connBuckets[GB_PAR_L] = g_connBuckets[GB_PAR_R] = NULL;
		g_vconnlist[0] = g_vconnlist[GB_PAR_L] = g_vconnlist[GB_PAR_R] = NULL;
		ConnBucketInitial(&(g_connBuckets[GB_PAR_L]), *g_maxvdeg + 1, &(g_vconnlist[GB_PAR_L]), *g_nv + 1);
		ConnBucketInitial(&(g_connBuckets[GB_PAR_R]), *g_maxvdeg + 1, &(g_vconnlist[GB_PAR_R]), *g_nv + 1);
		g_connMin[0] = g_connMax[0] = -1;
		g_connMin[GB_PAR_L] = g_connMin[GB_PAR_R] = *g_ne + 1;
		g_connMax[GB_PAR_L] = g_connMax[GB_PAR_R] = -1;
	}

	if (g_algo == GB_RTS || g_algo == GB_RRTS) {
		GainBucketInitial(&g_gainBuckets, 2 * (*g_maxvdeg) + 1, &g_vgainlist, *g_nv + 1);
		g_gainMin[0] = g_gainMax[0] = -1;
		g_gainMin[GB_PAR_L] = g_gainMin[GB_PAR_R] = *g_maxvdeg;
		g_gainMax[GB_PAR_L] = g_gainMax[GB_PAR_R] = 0 - *g_maxvdeg;
		g_gainBktsOffset = *g_maxvdeg;
		g_lastusedlist = (integer *) malloc((*g_nv + 1) * sizeof(integer));
		if (g_lastusedlist == NULL) {
			GB_Error(GB_ERR_MEM, FUNCNAME);
		}
		for (i = 1; i <= *g_nv; i++) {
			g_lastusedlist[i] = 0 - *g_nv;
		}
		g_time_best = g_time = 0;
		if (g_iter >= IDVL_RATIO) {
			g_idvl = g_iter / IDVL_RATIO;
		} else {
			g_idvl = 1;
		}
	}

	if (g_algo == GB_RRTS) {
		g_elite[0].edgecut = 0;
		g_elite[0].assign = NULL;
		for (i = 1; i <= TABU_NUM; i++) {
			g_elite[i].edgecut = *g_edgecut;
			g_elite[i].assign = (integer *) malloc((*g_nv + 1) * sizeof(integer));
			if (g_elite[i].assign == NULL) {
				GB_Error(GB_ERR_MEM, FUNCNAME);
			}
		}
	}

	/* Set evil seed */
	srand((unsigned) time(NULL));
}

void GB_GlobalFinalized() {
	integer i;

	if (g_algo == GB_MMG || g_algo == GB_RTS || g_algo == GB_RRTS) {
		ConnBucketFinalize(&(g_connBuckets[GB_PAR_L]), &(g_vconnlist[GB_PAR_L]), *g_nv + 1);
		ConnBucketFinalize(&(g_connBuckets[GB_PAR_R]), &(g_vconnlist[GB_PAR_R]), *g_nv + 1);
	}
	if (g_algo == GB_RTS || g_algo == GB_RRTS) {
		GainBucketFinalize(&g_gainBuckets, &g_vgainlist, *g_nv + 1);
		if (g_lastusedlist != NULL) {
			free(g_lastusedlist);
		}
	}
	if (g_algo == GB_RRTS) {
		for (i = 0; i <= TABU_NUM; i++) {
			if (g_elite[i].assign != NULL) {
				free(g_elite[i].assign);
			}
		}
	}
	if (g_algo == GB_GG) {
		GainBucketFinalize(&g_gainBuckets, &g_vgainlist, *g_nv + 1);
		if (g_vcutlist != NULL) {
			free(g_vcutlist);
		}
	}
	if (g_algo == GB_LKL || g_algo == GB_RRTS) {
		if (g_vlocklist != NULL) {
			free(g_vlocklist);
		}
		if (g_edgecutlist != NULL) {
			free(g_edgecutlist);
		}
		if (g_vpairlist != NULL) {
			free(g_vpairlist);
		}
		GainBucketFinalize(&(g_gainBkt[GB_PAR_L]), &(g_gainlst[GB_PAR_L]), *g_nv + 1);
		GainBucketFinalize(&(g_gainBkt[GB_PAR_R]), &(g_gainlst[GB_PAR_R]), *g_nv + 1);
	}
	if (g_algo == GB_KL) {
		if (g_dvaluelist != NULL) {
			free(g_dvaluelist);
		}
		if (g_vlocklist != NULL) {
			free(g_vlocklist);
		}
		if (g_edgecutlist != NULL) {
			free(g_edgecutlist);
		}
		if (g_vpairlist != NULL) {
			free(g_vpairlist);
		}
	}
	
	if (g_randomvlist != NULL) {
		free(g_randomvlist);
	}
	if (g_bestbelonglist != NULL) {
		free(g_bestbelonglist);
	}

}

