/* 
 * Graph Bisect Library
 * Randomized Tabu Search
 * Reactive Randomized Tabu Search
 *
 * Grid Challenge 2006 Final
 * Team Leo - Nan DUN
 * Department of Computer Science, the University of Tokyo
 */

#include "gbimpl.h"

bucket_t *g_gainBuckets;
integer g_gainMin[3], g_gainMax[3], g_gainBktsOffset;
itemp_t *g_vgainlist;
integer g_idvl, g_time, g_time_best, *g_lastusedlist;
double g_vote[TABU_NUM + 1];
elite_t g_elite[TABU_NUM + 1];

extern bucket_t *g_connBuckets[3];
extern integer g_connMin[3], g_connMax[3];
extern itemp_t *g_vconnlist[3];

void GainBucketInitial(bucket_t **buckets, integer ng, itemp_t **vlist, integer ni);
void GainBucketFinalize(bucket_t **buckets, itemp_t **vlist, integer ni);
void GainBucketReset(bucket_t *buckets, integer ng, itemp_t *vlist, integer ni);
void GainBucketInsert(int set, integer gain, integer vertex, bucket_t *buckets, integer *gainMin, integer *gainMax, integer offset, itemp_t *vlist);
void GainBucketDelete(int set, integer gain, integer vertex, bucket_t *buckets, integer *gainMin, integer *gainMax, integer offset, itemp_t *vlist);
void GainBucketUpdate(int oldset, int newset, integer oldgain, integer newgain, integer vertex, bucket_t *buckets, integer *gainMin, integer *gainMax, integer offset, itemp_t *vlist);
integer BestMove(int set, integer tabu);
void FixedTabuSearch(double tabufrac, int iter);
double Scoring(int trials);

void GB_ReactiveRandomizedTSBisect() {
	double tabufrac, besttabufrac;
	int time_start, time_idvl, rand, randrange;
	int i, j, k, elitecount;

	g_timelimit = g_timeout + g_bi_start;

	besttabufrac = Scoring(SCORING_TRIALS);
	if (g_exitflag == GB_TRUE) {
		goto FOUND_EXIT;
	}

	randrange = 0;
	for (i = 1; i <= TABU_NUM; i++) {
		randrange += (int) (g_vote[i] * VOTE_RATIO);	
	}
	if (randrange == 0) {
		randrange = 1;
	}
	elitecount = 1;
	for (i = 1; i <= (g_iter/g_idvl + 1); i++) {
		if (elitecount <= TABU_NUM) {
			for (j = elitecount; j <= TABU_NUM; j++) {
				if (g_elite[j].assign != NULL) {
					*g_edgecut = g_elite[j].edgecut;
					for (k = 1; k <= *g_nv; k++) {
						(*g_belonglist)[k] = (g_elite[j].assign)[k];
					}
					break;
				}
			}
			elitecount++;
		}
		else {
			*g_edgecut = *g_ne;
			for (k = 1; k <= *g_nv; k++) {
				(*g_belonglist)[k] = GB_PAR_L;
			}
			GB_MinMaxGreedyBisectSub();
				if (g_exitflag == GB_TRUE) {
					goto FOUND_EXIT;
				}
		}
		tabufrac = besttabufrac;
		time_idvl = g_time;
		do {
			time_start = g_time;
			do {
				FixedTabuSearch(tabufrac, 2 * ((integer)(tabufrac * (*g_nv)) + 1));
				if (g_exitflag == GB_TRUE) {
					goto FOUND_EXIT;
				}
				/* LOCAL-SEARCH */
#ifdef	LOCAL_SEARCH
				for (k = 1; k <= LOCAL_SEARCH_NUM; k++) {
					GB_LinearKernighanLinBisectSub();
					if (g_exitflag == GB_TRUE) {
						goto FOUND_EXIT;
					}
				}
#endif
			} while (g_time < time_start + *g_nv);
			if (g_time_best <= time_start) {
				rand = GB_Random(1, randrange);
				for (k = 1; k <= TABU_NUM; k++) {
					if ((rand > g_vote[k - 1] * VOTE_RATIO) && (rand <= g_vote[k] * VOTE_RATIO)) {
						tabufrac = k * 0.01;
						break;
					}
				}
			}
			if (g_bestedgecut <= g_threshold || (g_timeout > 0 && clock() >= g_timelimit)) {
				goto FOUND_EXIT;
			}
		} while (g_time < time_idvl + g_idvl);
	
	}
	
FOUND_EXIT:
	*g_edgecut = g_bestedgecut;
	for (i = 1; i <= *g_nv; i++) {
		(*g_belonglist)[i] = g_bestbelonglist[i];	
	}
}

double Scoring(int trials) {
	integer edgecut_start, n_2;
	double tabufrac, besttabufrac, votemin, votemax;
	int time_start, i, j, k;
	
	besttabufrac = tabufrac = 0.01;
	n_2 = *g_nv / 2;
	votemin = (double) *g_ne;
	votemax = 0.0;
	for (i = 1; i <= TABU_NUM; i++) {
		tabufrac = 0.01 * i;
		g_vote[i] = 0.0;
		for (j = 1; j <= trials; j++) {
			for (k = 1; k <= *g_nv; k++) {
				(*g_belonglist)[k] = GB_PAR_L;
			}
			GB_MinMaxGreedyBisectSub();
			if (g_exitflag == GB_TRUE) {
				return 0.0;
			}
			/* LOCAL-SEARCH */
#ifdef	LOCAL_SEARCH
			for (k = 1; k <= LOCAL_SEARCH_NUM; k++) {
				GB_LinearKernighanLinBisectSub();
				if (g_exitflag == GB_TRUE) {
					return 0.0;
				}
			}
#endif
			edgecut_start = *g_edgecut;
			time_start = g_time;
			do {
				FixedTabuSearch(tabufrac, 2 * ((integer)(tabufrac * (*g_nv)) + 1));
				if (g_exitflag == GB_TRUE) {
					return 0.0;
				}
				/* LOCAL-SEARCH */
#ifdef	LOCAL_SEARCH
				for (k = 1; k <= LOCAL_SEARCH_NUM; k++) {
					GB_LinearKernighanLinBisectSub();
					if (g_exitflag == GB_TRUE) {
						return 0.0;
					}
				}
#endif
			} while (g_time < time_start + n_2);

			if (g_bestedgecut == 0) {	/* Perfect bisection, return immediately */
				return 0.0;
			}
			if (g_elite[i].edgecut > *g_edgecut) {
				g_elite[i].edgecut = *g_edgecut;
				for (k = 1; k <= *g_nv; k++) {
					(g_elite[i].assign)[k] = (*g_belonglist)[k];
				}
			}
			//if (edgecut_start >= *g_edgecut) {
				g_vote[i] += ((double)(edgecut_start - *g_edgecut)) / (g_time - time_start);
			//}
			if (g_vote[i] < votemin) {
				votemin = g_vote[i];
			}
			if (g_vote[i] > votemax) {
				votemax = g_vote[i];
			}
		}
	}

	if (votemax != votemin) {
		for (i = 1; i <= TABU_NUM; i++) {
			g_vote[i] = 0.1 + 0.9 * (g_vote[i] - votemin) / (votemax - votemin);
		}
	}
	
	besttabufrac = tabufrac;
	for (i = 1; i <= TABU_NUM; i++) {
		if (g_vote[i] == votemax) {
			besttabufrac = 0.01 * i;
			break;
		}
	}
	return besttabufrac;
}

void GB_RandomizedTabuSearchBisect() {
	int iterend, i;
	double tabufrac;
	clock_t timeout;

	timeout = g_timeout + g_bi_start;
	while (g_time < g_iter) {
		for (i = 1; i <= *g_nv; i++) {
			(*g_belonglist)[i] = GB_PAR_L; 
		}
		GB_MinMaxGreedyBisectSub();
		iterend = g_time + g_idvl;
		while (g_time < iterend) {
			tabufrac = 0.01 * GB_Random(1, 25);
			FixedTabuSearch(tabufrac, *g_nv);
			if (g_bestedgecut == 0 || (g_timeout > 0 && clock() >= timeout)) {	/* perfect bisection, don't waste time */
				goto FOUND_EXIT;
			}
		}
	}

FOUND_EXIT:
	*g_edgecut = g_bestedgecut;
	for (i = 1; i <= *g_nv; i++) {
		(*g_belonglist)[i] = g_bestbelonglist[i];	
	}
}

void FixedTabuSearch(double tabufrac, int iter) {
	const char FUNCNAME[] = "FixedTabuSearch(2)";

	integer v, tabu, n[3], n_2;
	integer seed,	i, j, tempvertex, tempset;
	int addset, otherset, gain, oldgain, newgain;

	if (g_gainBuckets == NULL || g_vgainlist == NULL || g_randomvlist == NULL) {
		GB_Error(GB_ERR_NULL, FUNCNAME);
	}

	/* Start Initialization */
	GainBucketReset(g_gainBuckets, 2 * (*g_maxvdeg) + 1, g_vgainlist, *g_nv + 1);
	g_gainMin[GB_PAR_L] = g_gainMin[GB_PAR_R] = *g_maxvdeg;
	g_gainMax[GB_PAR_L] = g_gainMax[GB_PAR_R] = 0 - *g_maxvdeg;
	GB_RandomPermute(g_randomvlist, *g_nv);
	/* End Initialization */
	
	/* Setup gainBkt[L_PAR] and gainBkt[R_PAR] */
	for (i = 1; i <= *g_nv; i++) {
		seed = g_randomvlist[i];
		addset = (*g_belonglist)[seed];
		gain = 0;
		for (j = 1; j <= (*g_adjlist)[seed][0]; j++) {
			tempvertex = (*g_adjlist)[seed][j];
			if (addset == (*g_belonglist)[tempvertex]) {
				gain++;
			}
			else {
				gain--;
			}
		}
		GainBucketInsert(addset, gain, seed, g_gainBuckets, g_gainMin, g_gainMax, g_gainBktsOffset, g_vgainlist);
	}

	/* Start Fixed Tabu Search */
	tabu = (integer) (tabufrac * (*g_nv));
	if (*g_vparity == GB_NUM_EVEN) {
		n[GB_PAR_L] = n[GB_PAR_R] = n_2 = *g_nv / 2;
	} else {
		n[GB_PAR_L] = n[GB_PAR_R] = 0;
		for (i = 1; i <= *g_nv; i++) {
			n[(*g_belonglist)[i]]++;
		}
		if (n[GB_PAR_L] % 2 == 0) {
			n_2 = n[GB_PAR_L];
		} else {
			n_2 = n[GB_PAR_R];
		}
	}
	for (i = 1; i <= iter; i++)	{
		if (n[GB_PAR_L] >= n_2) {
			addset = GB_PAR_R;
			otherset = GB_PAR_L;
		}
		else {
			addset = GB_PAR_L;
			otherset = GB_PAR_R;
		}
        v = BestMove(otherset, tabu);
		oldgain = g_vgainlist[v].value - g_gainBktsOffset;
		*g_edgecut += oldgain;
		(*g_belonglist)[v] = addset;
		n[addset] += 1;
		n[otherset] -= 1;

		newgain = 0;
		for (j = 1; j <= (*g_adjlist)[v][0]; j++)
		{
			tempvertex = (*g_adjlist)[v][j];
			if (addset == (*g_belonglist)[tempvertex]) {
				newgain++;
			}
			else {
				newgain--;
			}
		}
		GainBucketUpdate(otherset, addset, oldgain, newgain, v, g_gainBuckets, g_gainMin, g_gainMax, g_gainBktsOffset, g_vgainlist);
		/* update v's adjacent node */
		for (j = 1; j <= (*g_adjlist)[v][0]; j++)
		{
			tempvertex = (*g_adjlist)[v][j];
			tempset = (*g_belonglist)[tempvertex];
			oldgain = g_vgainlist[tempvertex].value - g_gainBktsOffset;
			if ((*g_belonglist)[v] == tempset) {
				newgain = oldgain + 2;
			}
			else {
				newgain = oldgain - 2;
			}
			GainBucketUpdate(tempset, tempset, oldgain, newgain, tempvertex, g_gainBuckets, g_gainMin, g_gainMax, g_gainBktsOffset, g_vgainlist);
		}
		g_lastusedlist[v] = g_time;
		g_time++;
		if (*g_vparity == GB_NUM_EVEN) {
			if (n[GB_PAR_L] == n[GB_PAR_R] && (*g_edgecut < g_bestedgecut)) {
				g_bestedgecut = *g_edgecut;
				for (j = 1; j <= *g_nv; j++) {
					g_bestbelonglist[j] = (*g_belonglist)[j];					
				}
				g_time_best = g_time;
			}
		} else {
			if (((n[GB_PAR_L] - n[GB_PAR_R]) == 1 || (n[GB_PAR_R] - n[GB_PAR_L]) == 1) && (*g_edgecut < g_bestedgecut)) {
				g_bestedgecut = *g_edgecut;
				for (j = 1; j <= *g_nv; j++) {
					g_bestbelonglist[j] = (*g_belonglist)[j];
				}
				g_time_best = g_time;
			}
		}
		if (g_bestedgecut <= g_threshold || (g_timeout > 0 && clock() >= g_timelimit)) {
			g_exitflag = GB_TRUE;
			return;
		}
	}

}

integer BestMove(int set, integer tabu) {
	integer i, v, floor, celling;
	item_t *curr;
	v = 0;
	floor = g_gainMin[set] + g_gainBktsOffset;
	celling = g_gainMax[set] + g_gainBktsOffset;
	for (i = floor; i <= celling; i++) {
		curr = g_gainBuckets[i].head;
		while (curr != NULL) {
			v = curr->index;
			if ((g_lastusedlist[v] < (g_time - tabu)) && ((*g_belonglist)[v] == set)) {
				goto FOUND_EXIT;
			}
			curr = curr->next;
		}
	}
FOUND_EXIT:
	return v;
}

void GainBucketInitial(bucket_t **buckets, integer ng, itemp_t **vlist, integer ni) {
	GB_BucketInitial(buckets, ng, vlist, ni, 0);
}

void GainBucketFinalize(bucket_t **buckets, itemp_t **vlist, integer ni) {
	GB_BucketFinalize(buckets, vlist, ni);
}

void GainBucketReset(bucket_t *buckets, integer ng, itemp_t *vlist, integer ni) {
	GB_BucketReset(buckets, ng, vlist, ni, 0);
}

void GainBucketInsert(int set, integer gain, integer vertex, bucket_t *buckets, integer *gainMin, integer *gainMax, integer offset, itemp_t *vlist) {
	integer mygain, mymin, mymax;
	mygain = gain + offset;
	mymin = gainMin[set] + offset;
	mymax = gainMax[set] + offset;
	GB_BucketInsert(mygain, vertex, buckets, &mymin, &mymax, vlist);
	gainMin[set] = mymin - offset;
	gainMax[set] = mymax - offset;
}

void GainBucketDelete(int set, integer gain, integer vertex, bucket_t *buckets, integer *gainMin, integer *gainMax, integer offset, itemp_t *vlist) {
	integer mygain, mymin, mymax;

	mygain = gain + offset;
	mymin = gainMin[set] + offset;
	mymax = gainMax[set] + offset;

	GB_BucketDelete(mygain, vertex, buckets, &mymin, &mymax, vlist);
	
	if (mymin > mymax) {
		gainMin[set] = *g_maxvdeg;
		gainMax[set] = 0 - *g_maxvdeg;
	} else {
		gainMin[set] = mymin - offset;
		gainMax[set] = mymax - offset;
	}
}

void GainBucketUpdate(int oldset, int newset, integer oldgain, integer newgain, integer vertex, bucket_t *buckets, integer *gainMin, integer *gainMax, integer offset, itemp_t *vlist) {
	integer mygain, mymin, mymax;

	mygain = oldgain + offset;
	mymin = gainMin[oldset] + offset;
	mymax = gainMax[oldset] + offset;

	GB_BucketDelete(mygain, vertex, buckets, &mymin, &mymax, vlist);

	if (oldset != newset) {
		gainMin[oldset] = mymin - offset;
		gainMax[oldset] = mymax - offset;
		mymin = gainMin[newset] + offset;
		mymax = gainMax[newset] + offset;
	}
	mygain = newgain + offset;

	GB_BucketInsert(mygain, vertex, buckets, &mymin, &mymax, vlist);

	gainMin[newset] = mymin - offset;
	gainMax[newset] = mymax - offset;
}

